#include "box.hh"
#include "common.hh"
#include "cylinder.hh"
#include "ground.hh"
#include "object.hh"
#include "params.hh"
#include "vector.hh"
#include "vertex.hh"
#include <iostream>
#include <cstdlib>

using namespace std;

slist<Object*> world;

Vertex cloth[X_SIZE][Y_SIZE];
Vector normals[X_SIZE][Y_SIZE];

void output_cyl();
void compute_normals();

// transform the cloth:
void translate (const Vector&);
void rotate (const Vector&);

void createWorld() {
  //world.push_front (new GroundPlane (-0.1));

  world.push_front (new Box (Vector (-10, -21.5, -50), Vector (10, 21.5, 0.01)));
  world.push_front (new GroundPlane (-48.9999));
  translate ((-X_SIZE*dist/2.0)*X);
  rotate ((PI/3)*Y);
  translate (10*X);
  rotate ((-PI/6)*Z);
  translate (-25*Y);

//  world.push_front (new Cylinder (-3*Z, -Z, 43));

//  world.push_front (new GroundPlane (-50));
//  world.push_front (new Cylinder (-60*Z+25*X, -10*Z+30*X, 2.5));
//  world.push_front (new Cylinder (-60*Z-25*X+25*Y, -5*Z-30*X+30*Y, 2.5));
//  world.push_front (new Cylinder (-60*Z-25*X-25*Y, -Z-30*X-30*Y, 2.5));
}

void translate (const Vector& v) {
  for (int i = 0; i < X_SIZE; i++)
    for (int j = 0; j < Y_SIZE; j++)
      cloth[i][j].location_ += v;
}

void rotate (const Vector& v) {
  Vector coefs[3];
  double cx = cos (v.x()), cy = cos (v.y()), cz = cos (v.z());
  double sx = sin (v.x()), sy = sin (v.y()), sz = sin (v.z());


  coefs[0] = Vector (cy*cz, sx*sy*cz-cx*sz, cx*sy*cz+sx*sz);
  coefs[1] = Vector (cy*sz, sx*sy*sz+cx*cz, cx*sy*sz-sx*cz);
  coefs[2] = Vector (  -sy,          sx*cy,          cx*cy);

  for (int i = 0; i < X_SIZE; i++)
    for (int j = 0; j < Y_SIZE; j++)
      cloth[i][j].location_ = Vector (cloth[i][j].location_.dot (coefs[0]),
                                      cloth[i][j].location_.dot (coefs[1]),
                                      cloth[i][j].location_.dot (coefs[2]));
}

void init() {
  for (int i = 0; i < X_SIZE; i++)
    for (int j = 0; j < Y_SIZE; j++)
      cloth[i][j] = Vertex (Vector ((i-X_SIZE/2)*dist*(1-HNOISE-VNOISE)+(random()/(double)RAND_MAX)*dist*HNOISE, 
                                    (j-Y_SIZE/2)*dist*(1-HNOISE-VNOISE)+(random()/(double)RAND_MAX)*dist*HNOISE,
                                    (random()/(double)RAND_MAX)*VNOISE));
  // I still need to link them to their neighbors...
  // neighbors
  for (int i = 0; i < X_SIZE-1; i++)
    for (int j = 0; j < Y_SIZE-1; j++) {
      cloth[i][j].addNeighbor (&cloth[i+1][j]);
      cloth[i][j].addNeighbor (&cloth[i][j+1]);
    }
  for (int i = 0; i < X_SIZE-1; i++)
    cloth[i][Y_SIZE-1].addNeighbor (&cloth[i+1][Y_SIZE-1]);
  // neighbors and corners
  for (int j = 0; j < Y_SIZE-1; j++) {
    cloth[X_SIZE-1][j].addNeighbor (&cloth[X_SIZE-1][j+1]);
    cloth[X_SIZE-1][j].addCorner (&cloth[X_SIZE-2][j+1]);
  // corners
    cloth[0][j].addCorner (&cloth[1][j+1]);
  }
  for (int i = 1; i < X_SIZE-1; i++)
    for (int j = 0; j < Y_SIZE-1; j++) {
      cloth[i][j].addCorner (&cloth[i-1][j+1]);
      cloth[i][j].addCorner (&cloth[i+1][j+1]);
    }
  // curvings
  for (int i = 1; i < X_SIZE-1; i++) {
    cloth[i][0].addCurving (&cloth[i-1][0], &cloth[i+1][0]);
    cloth[i][Y_SIZE-1].addCurving (&cloth[i-1][Y_SIZE-1], &cloth[i+1][Y_SIZE-1]);
  }
  for (int j = 1; j < Y_SIZE-1; j++) {
    cloth[0][j].addCurving (&cloth[0][j-1], &cloth[0][j+1]);
    cloth[X_SIZE-1][j].addCurving (&cloth[X_SIZE-1][j-1], &cloth[X_SIZE-1][j+1]);
  }
  for (int i = 1; i < X_SIZE-1; i++)
    for (int j = 1; j < Y_SIZE-1; j++) {
      cloth[i][j].addCurving (&cloth[i-1][j], &cloth[i+1][j]);
      cloth[i][j].addCurving (&cloth[i][j-1], &cloth[i][j+1]);
    }
}

void compute_forces() {
  for (int i = 0; i < X_SIZE; i++)
    for (int j = 0; j < Y_SIZE; j++)
      cloth[i][j].computeForce();
}

void move() {
  for (int i = 0; i < X_SIZE; i++)
    for (int j = 0; j < Y_SIZE; j++) {
      cloth[i][j].move();
      cloth[i][j].doMove();
      cloth[i][j].reset();
    }
}

#define CLOTH(X_, Y_) cloth[X_][Y_].location()

void output_cyl() {
  cout << "union {" << endl;
  for (int i = 1; i < X_SIZE; i++)
    for (int j = 1; j < Y_SIZE; j++) {
      cout << " cylinder { " << CLOTH(i-1, j) << ", " << CLOTH(i, j) << ", " << dist/10.0 << " pigment { red 1 } }" << endl;
      cout << " cylinder { " << CLOTH(i, j-1) << ", " << CLOTH(i, j) << ", " << dist/10.0 << " pigment { green 1 } }" << endl;
    }
  cout << endl << "}" << endl;
}

void output() {
  cout << "mesh {" << endl;
  for (int i = 1; i < X_SIZE; i++) {
    for (int j = 1; j < Y_SIZE; j++) {
      cout << "  smooth_triangle { " 
           << CLOTH(i, j) << ", " << normals[i][j] << ", "
           << CLOTH(i-1, j) << ", " << normals[i-1][j] << ", "
           << CLOTH(i-1, j-1) << ", " << normals[i-1][j-1]
           << " }" << endl;
      cout << "  smooth_triangle { "
           << CLOTH(i, j) << ", " << normals[i][j] << ", "
           << CLOTH(i, j-1) << ", " << normals[i][j-1] << ", "
           << CLOTH(i-1, j-1) << ", " << normals[i-1][j-1]
           << " }" << endl;
      cout << endl;
    }
    cout << "/******************************/" << endl;
  }
  cout << endl << "}" << endl;
}

void output2() {
  cout << "#version unofficial MegaPov 0.6;" << endl;
  cout << "mesh2 {" << endl;
  cout << "  vertex_vectors {" << endl << "    " << X_SIZE*Y_SIZE;
  for (int i = 0; i < X_SIZE; i++)
    for (int j = 0; j < Y_SIZE; j++)
      cout << "," << endl << "    " << CLOTH(i, j);
  cout << endl << "  }" << endl;
  cout << "  normal_vectors {" << endl << "    " << X_SIZE*Y_SIZE;
  for (int i = 0; i < X_SIZE; i++)
    for (int j = 0; j < Y_SIZE; j++)
      cout << "," << endl << "    " << normals[i][j];
  cout << endl << "  }" << endl;
  cout << "  uv_vectors {" << endl << "    " << X_SIZE*Y_SIZE;
  for (int i = 0; i < X_SIZE; i++)
    for (int j = 0; j < Y_SIZE; j++)
      cout << "," << endl << "    <" << i/(double)(X_SIZE-1) << ", " << j/(double)(Y_SIZE-1) << ">";
  cout << endl << "  }" << endl;
  cout << "  face_indices {" << endl << "    " << 2*(X_SIZE-1)*(Y_SIZE-1);
  for (int i = 1; i < X_SIZE; i++)
    for (int j = 1; j < Y_SIZE; j++) {
      cout << "," << endl;
      cout << "    <" << i+j*Y_SIZE << ", " << i-1+j*Y_SIZE << ", " << i-1+(j-1)*Y_SIZE << ">, " << endl;
      cout << "    <" << i+j*Y_SIZE << ", " << i+(j-1)*Y_SIZE << ", " << i-1+(j-1)*Y_SIZE << ">";
    }
  cout << endl << "  }" << endl;
  cout << endl << "}" << endl;
}

void compute_normals() {
                                // corners
  normals[0][0] = (1.0/(dist*dist))*(CLOTH(1, 0)-CLOTH(0, 0))
    .cross (CLOTH(0, 1)-CLOTH(0, 0));
  normals[0][Y_SIZE-1] = (1.0/(dist*dist))*(CLOTH(1, Y_SIZE-1)-CLOTH(0, Y_SIZE-1))
    .cross (CLOTH(0, Y_SIZE-1)-CLOTH(0, Y_SIZE-2));
  normals[X_SIZE-1][0] = (1.0/(dist*dist))*(CLOTH(X_SIZE-1, 0)-CLOTH(X_SIZE-2, 0))
    .cross (CLOTH(X_SIZE-1, 1)-CLOTH(X_SIZE-1, 0));
  normals[X_SIZE-1][Y_SIZE-1] = (1.0/(dist*dist))*(CLOTH(X_SIZE-1, Y_SIZE-1)-CLOTH(X_SIZE-2, Y_SIZE-1))
    .cross (CLOTH(X_SIZE-1, Y_SIZE-1)-CLOTH(X_SIZE-1, Y_SIZE-2));
                                // sides
  for (int j = 1; j < Y_SIZE-1; j++) {
    normals[0][j] = (1.0/(2*dist*dist))*(CLOTH(1, j)-CLOTH(0, j))
      .cross (CLOTH(0, j+1)-CLOTH(0, j-1));
    normals[X_SIZE-1][j] = (1.0/(2*dist*dist))*(CLOTH(X_SIZE-1, j)-CLOTH(X_SIZE-2, j))
      .cross (CLOTH(X_SIZE-1, j+1)-CLOTH(X_SIZE-1, j-1));
  }
  for (int i = 1; i < X_SIZE-1; i++) {
    normals[i][0] = (1.0/(2*dist*dist))*(CLOTH(i+1, 0)-CLOTH(i-1, 0))
      .cross (CLOTH(i, 1)-CLOTH(i, 0));
    normals[i][Y_SIZE-1] = (1.0/(2*dist*dist))*(CLOTH(i+1, Y_SIZE-1)-CLOTH(i-1, Y_SIZE-1))
      .cross (CLOTH(i, Y_SIZE-1)-CLOTH(i, Y_SIZE-2));
  }
                                // middle
  for (int i = 1; i < X_SIZE-1; i++)
    for (int j = 1; j < Y_SIZE-1; j++)
      normals[i][j] = (1.0/(4*dist*dist))*(CLOTH(i+1, j)-CLOTH(i-1, j))
        .cross (CLOTH(i, j+1)-CLOTH(i, j-1));
}

int main() {
  init();
  createWorld();
  for (int i = 0; i < STEPS; i++) {
    compute_forces();
    move();
    if ((i & 0xF) == 0)
      if ((i & 0x3FF) != 0)
        cerr << ".";
      else
        cerr << " " << i << endl;
  }
  cerr << endl;

  compute_forces();
  bool ok = false;
  for (int i = 0; i < X_SIZE; i++) {
    for (int j = 0; j < Y_SIZE; j++) {
      cloth[i][j].move();
      cloth[i][j].doMove();
      ok |= cloth[i][j].moving();
    }
  }
  if (ok)
    cerr << "WARNING: inclomplete computation!" << endl;
  compute_normals();
  cout.setf (ios::showpoint);
  cout.setf (ios::showpos);
  cout.setf (ios::scientific, ios::floatfield);
  output2();
  return 0;
}
