#include "common.hh"
#include "object.hh"
#include "params.hh"
#include "vertex.hh"

#include <slist>

Vertex::Vertex(): location_ (), force(), newloc(), neighbors(),
                  corners() {}

Vertex::Vertex (const Vector& iloc): location_ (iloc), force(),
                                     newloc(), neighbors(), 
                                     corners() {}

void Vertex::addNeighbor (Vertex* n) {
  neighbors.push_front (n);
}

void Vertex::addCorner (Vertex* c) {
  corners.push_front (c);
}

void Vertex::addCurving (Vertex* first, Vertex* second) {
  curvings.push_front (pair<Vertex*, Vertex*> (first, second));
}

void Vertex::computeForce() {
  // weight
  force -= weight*Z;

  // neighbors
  for (slist<Vertex*>::iterator n = neighbors.begin(); n != neighbors.end(); n++) {
    Vector v = location_ - (*n)->location_;
    double l = v.length();
    Vector f = ((l - dist)*strength/l)*v;
    force -= f;
    (*n)->force += f;
  }

  // corners
  for (slist<Vertex*>::iterator c = corners.begin(); c != corners.end(); c++) {
    Vector v = location_ - (*c)->location_;
    double l = v.length();
    Vector f = ((l - d2)*s2/l)*v;
    force -= f;
    (*c)->force += f;
  }

  // curving
  if (!curvings.empty()) {
    for (slist<pair<Vertex*, Vertex*> >::iterator f = curvings.begin();
         f != curvings.end(); f++) {
      Vector v = (f->first->location_ - f->second->location_);
      v *= (curve*(2*dist-v.length()));
      f->first->force += v;
      f->second->force -= v;
    }
  }
}

void Vertex::adjust() {
  for (slist<Object*>::iterator o = world.begin(); o != world.end(); o++)
    (*o)->adjust (force, location_);
}

void Vertex::move() {
  adjust();
  newloc = location_ + step*force;
//  force = O;
}

void Vertex::doMove() {
  for (slist<Object*>::iterator o = world.begin(); o != world.end(); o++)
    (*o)->inside (newloc, location_);
  location_ = newloc;
}
